import re
from abc import ABCMeta, abstractmethod
from collections import defaultdict

from .status_codes import DictStatusCode, PERMANENT_NEGATIVE_COMPLETION_CODES


class BaseResponse(metaclass=ABCMeta):
    STATUS_REGEXP = re.compile(r"^\d{3}")
    CONTENT_DELIMITER = "."
    LINE_DELIMITER = "\r\n"

    def __init__(self, response_bytes):
        self.response_text = response_bytes.decode()
        self.status_code = self.parse_status_code()
        self.content = self.parse_content()

    def parse_status_code(self):
        match = self.STATUS_REGEXP.match(self.response_text)
        if not match:
            raise ValueError(
                f"Expected status response but received: {self.response_text}"
            )
        return int(match.group(0))

    @abstractmethod
    def parse_content(self):
        pass

    def get_multipart_content_lines(self):
        return list(
            filter(self.is_content_line, self.response_text.split(self.LINE_DELIMITER))
        )

    def is_content_line(self, line):
        return self.STATUS_REGEXP.match(line) is None


class ServerPropertiesResponse(BaseResponse):
    """ Responses to a SHOW DB or SHOW STRAT command
    """

    def parse_content(self):
        if self.status_code in PERMANENT_NEGATIVE_COMPLETION_CODES:
            return None
        response_lines = self.response_text.split(self.LINE_DELIMITER)
        content_lines = list(filter(self.is_content_line, response_lines))
        content_lines = content_lines[: content_lines.index(self.CONTENT_DELIMITER)]
        lines_split = (line.split(maxsplit=1) for line in content_lines)
        return {item: description.strip('"') for item, description in lines_split}


class PreliminaryResponse(BaseResponse):
    """ A generic one line response (status code followed by optional
    text).
    """

    def parse_content(self):
        return self.response_text.split(maxsplit=1)[1].strip()


class DefineWordResponse(BaseResponse):
    def parse_content(self):
        if self.status_code == DictStatusCode.NO_MATCH:
            return None
        definition_lines = self.get_multipart_content_lines()
        definitions = []
        while "." in definition_lines:
            delim_index = definition_lines.index(self.CONTENT_DELIMITER)
            new_def = definition_lines[:delim_index]
            definitions.append(
                {"db": new_def[0].split()[2], "definition": "\n".join(new_def[1:])}
            )
            definition_lines = definition_lines[delim_index + 1 :]
        return definitions

    def is_content_line(self, line):
        return not line.startswith("150") and not line.startswith("250")


class MatchResponse(BaseResponse):
    def parse_content(self):
        if self.status_code == DictStatusCode.NO_MATCH:
            return None
        match_lines = self.get_multipart_content_lines()
        match_lines = match_lines[: match_lines.index(self.CONTENT_DELIMITER)]
        matches = defaultdict(list)
        for line in match_lines:
            db_name, match = line.split(maxsplit=1)
            matches[db_name].append(match.strip('"'))
        return matches


class MultiLineResponse(BaseResponse):
    def parse_content(self):
        content_lines = self.get_multipart_content_lines()
        content_lines = content_lines[: content_lines.index(self.CONTENT_DELIMITER)]
        return "\n".join(content_lines)


class DatabaseInfoResponse(MultiLineResponse):
    def parse_content(self):
        if self.status_code == DictStatusCode.INVALID_DB:
            return None
        return super().parse_content()


class HandshakeResponse(PreliminaryResponse):
    CAPABILITIES_RE = re.compile(r"<([\w\d_]*(\.[\w\d_]+)*)>")
    MSG_ID_RE = re.compile(r"(<[\d\w@.]+>)\r\n")
    BANNER_RE = re.compile(r"^220 ([^<]+)?")

    def parse_content(self):
        capabilities_match = self.CAPABILITIES_RE.search(self.response_text)
        msg_id_match = self.MSG_ID_RE.search(self.response_text)
        banner_match = self.BANNER_RE.search(self.response_text)
        if not msg_id_match:
            raise ValueError(
                "Client got unexpected banner in connection response: "
                f"{self.response_text}"
            )
        content = {
            "message_id": msg_id_match.group(1),
            "capabilities": None,
            "banner": None,
        }
        if capabilities_match:
            content.update({"capabilities": capabilities_match.group(1).split(".")})
        if banner_match:
            content.update({"banner": banner_match.group(1).rstrip()})
        return content



import getpass
import select
import socket
from datetime import datetime

from .commands import (
    client_ident_command,
    define_word_command,
    disconnect_command,
    show_strategies_command,
    show_databases_command,
    match_command,
    status_command,
    show_info_command,
    show_server_command,
    help_command,
)
from .response import (
    HandshakeResponse,
    DefineWordResponse,
    PreliminaryResponse,
    ServerPropertiesResponse,
    MatchResponse,
    DatabaseInfoResponse,
    MultiLineResponse,
)
from .status_codes import DictStatusCode


BUF_SIZE = 4096
DEFAULT_PORT = 2628


class DictionaryClient:
    """Implements a client for communication with a server implementing
    the DICT Server Protocol (https://tools.ietf.org/html/rfc2229).
    """

    def __init__(self, host="localhost", port=DEFAULT_PORT, sock_class=socket.socket):
        self.client_name = f"{getpass.getuser()}@{socket.gethostname()}"
        self.client_id_info = f"{self.client_name} {datetime.now().isoformat()}"
        self.sock = sock_class(socket.AF_INET, socket.SOCK_STREAM)
        self.server_info = self._connect(host, port)
        self.strategies = self._get_strategies()
        self.databases = self._get_databases()

    def _recv_all(self):
        rlist, _, _ = select.select([self.sock], [], [], 5)
        if self.sock not in rlist:
            raise TimeoutError("Client timed out expecting server response.")
        bytes_received = self.sock.recv(BUF_SIZE)
        status_code = self._get_status(bytes_received)
        if DictStatusCode.response_complete(status_code):
            return bytes_received
        while not self._response_complete(bytes_received):
            rlist, _, _ = select.select([self.sock], [], [], 5)
            if self.sock not in rlist:
                raise TimeoutError(
                    "Client timed out following preliminary response with status "
                    f"{status_code}."
                )
            bytes_received += self.sock.recv(BUF_SIZE)
        return bytes_received

    def _connect(self, host, port):
        self.sock.connect((host, port))
        response = HandshakeResponse(self._recv_all())
        if response.status_code != DictStatusCode.CONNECTION_ACCEPTED:
            raise Exception(response.status_code)
        self._send_client_ident()
        return response.content

    def _send_client_ident(self):
        self.sock.sendall(client_ident_command(self.client_id_info))
        response = PreliminaryResponse(self._recv_all())
        if response.status_code != DictStatusCode.OK:
            raise Exception(response.status_code)

    def _get_status(self, response_bytes):
        return int(response_bytes[:3])

    def _response_complete(self, response_bytes):
        return b"250" in response_bytes and response_bytes[-2:] == b"\r\n"

    def _get_strategies(self):
        self.sock.sendall(show_strategies_command())
        response = ServerPropertiesResponse(self._recv_all())
        return response.content

    def _get_databases(self):
        self.sock.sendall(show_databases_command())
        response = ServerPropertiesResponse(self._recv_all())
        return response.content

    def _get_response(self, command, response_class):
        self.sock.sendall(command)
        return response_class(self._recv_all())

    def get_server_status(self):
        return self._get_response(status_command(), PreliminaryResponse)

    def get_server_information(self):
        return self._get_response(show_server_command(), MultiLineResponse)

    def get_db_info(self, db):
        if db not in self.databases:
            raise ValueError(f'Invalid database name: "{db}" not present.')
        return self._get_response(show_info_command(db), DatabaseInfoResponse)

    def get_help_text(self):
        return self._get_response(help_command(), MultiLineResponse)

    def define(self, word, db="*"):
        if db != "*" and db not in self.databases:
            raise ValueError(f'Invalid database name: "{db}" not present.')
        return self._get_response(define_word_command(word, db), DefineWordResponse)

    def match(self, word, db="*", strategy="."):
        if db != "*" and db not in self.databases:
            raise ValueError(f'Invalid database name: "{db}" not present.')
        if strategy != "." and strategy not in self.strategies:
            raise ValueError(f'Unknown strategy: "{strategy}".')
        return self._get_response(
            match_command(word, db=db, strategy=strategy), MatchResponse
        )

    def disconnect(self):
        self.sock.sendall(disconnect_command())
        bytes_recieved = self._recv_all()
        if self._get_status(bytes_recieved) != DictStatusCode.CLOSING_CONNECTION:
            raise ConnectionError(
                "Client got unexpected response to QUIT command: "
                f'"{bytes_recieved.decode()}"'
            )
        self.sock.close()



from functools import wraps


def dictionary_server_command(command_func):
    @wraps(command_func)
    def encodes_command(*args, **kwargs):
        command = command_func(*args, **kwargs)
        return f"{command}\r\n".encode("utf-8")

    return encodes_command


@dictionary_server_command
def show_strategies_command():
    return "SHOW STRAT"


@dictionary_server_command
def show_databases_command():
    return "SHOW DB"


@dictionary_server_command
def client_ident_command(client_id_info):
    return f"CLIENT {client_id_info}"


@dictionary_server_command
def disconnect_command():
    return "QUIT"


@dictionary_server_command
def define_word_command(word, db="*"):
    return f"DEFINE {db} {word}"


@dictionary_server_command
def match_command(word, db="*", strategy="exact"):
    return f"MATCH {db} {strategy} {word}"


@dictionary_server_command
def status_command():
    return "STATUS"


@dictionary_server_command
def show_info_command(db):
    return f"SHOW INFO {db}"


@dictionary_server_command
def show_server_command():
    return "SHOW SERVER"


@dictionary_server_command
def help_command():
    return "HELP"



from enum import IntEnum


class DictStatusCode(IntEnum):
    @staticmethod
    def response_complete(status):
        return status >= 200

    @staticmethod
    def text_follows(status):
        return status >= 100 and status < 200

    DATABASES_PRESENT = 110
    STRATEGIES_AVAILABLE = 111
    DB_INFO_FOLLOWS = 112
    HELP_TEXT_FOLLOWS = 113
    SERVER_INFO_FOLLOWS = 114
    SASL_CHALLENGE_FOLLOWS = 130
    DEFINITIONS_FOLLOW = 150
    DEFINITION_BODY = 151
    MATCHES_FOUND = 152

    STATUS_FOLLOWS = 210
    CONNECTION_ACCEPTED = 220
    CLOSING_CONNECTION = 221
    AUTH_SUCCESSFUL = 230
    OK = 250

    SASL_SEND_RESPONSE = 330

    SERVER_TEMP_UNAVAILABLE = 420
    SERVER_SHUTDOWN = 421

    CMD_NOT_RECOGNIZED = 500
    ILLEGAL_PARAMS = 501
    CMD_NOT_IMPLEMENTED = 502
    PARAM_NOT_IMPLEMENTED = 503
    ACCESS_DENIED = 530
    AUTH_FAILED = 531
    UNKNOWN_MECHANISM = 532
    INVALID_DB = 550
    INVALID_STRATEGY = 551
    NO_MATCH = 552
    NO_DATABASES = 554
    NO_STRATEGIES = 555


PERMANENT_NEGATIVE_COMPLETION_CODES = {
    DictStatusCode.CMD_NOT_RECOGNIZED,
    DictStatusCode.ILLEGAL_PARAMS,
    DictStatusCode.CMD_NOT_IMPLEMENTED,
    DictStatusCode.PARAM_NOT_IMPLEMENTED,
    DictStatusCode.ACCESS_DENIED,
    DictStatusCode.AUTH_FAILED,
    DictStatusCode.UNKNOWN_MECHANISM,
    DictStatusCode.INVALID_DB,
    DictStatusCode.INVALID_STRATEGY,
    DictStatusCode.NO_MATCH,
    DictStatusCode.NO_DATABASES,
    DictStatusCode.NO_STRATEGIES,
}



from os import path
from setuptools import setup

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="py-dict-client",
    version="0.1.4",
    description="A client implementing the Dictionary Server Protocol (DICT)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jams2/py-dict-client",
    author="Joshua Munn",
    author_email="public@elysee-munn.family",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
    ],
    keywords="dictionary language definition server client",
    package_dir={"dictionary_client": "dictionary_client"},
    packages=["dictionary_client"],
    python_requires=">=3",
)
